
# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-06-22T10:53:30+0100",
 "dirty": false,
 "error": null,
 "full-revisionid": "7c48ff4409c622c582c56a5702373f726de08e96",
 "version": "1.2.5"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
